#
# The Python Imaging Library.
#
# DM3 image file handling
#
# History:
# 2012-05-16    Created by Maofu Liao
#
## This calles another python script (dm3lib_v099.py) adapted by
## Pierre-Ivan Raynal <raynal@med.univ-tours.fr>
## http://microscopies.med.univ-tours.fr/
#
# ----------------------------------------------------------------------
#
# Author: Maofu Liao (maofuliao@gmail.com)
#
# This program is free open-source software, licensed under the terms of 
# the GNU Public License version 3 (GPLv3). If you redistribute it and/or 
# modify it, please make sure the contribution of Maofu Liao 
# is acknowledged appropriately.
#
#
#

import Image, ImageFile
import dm3lib_v099 as dm3


class Dm3ImageFile(ImageFile.ImageFile):

	format = "DM3"
	format_description = "DM3 image file"


	def _open(self):
		try:
			dimg = dm3.DM3(self.filename)
			# DataTypes for image data <--> PIL decoders
			dataTypes = {
				1: 'F;16S',    #16-bit LE signed integer
				2: 'F;32F',
				7: 'F;32',   #32-bit LE unsigned integer
				10: 'F;16',    #16-bit LE unsigned integer
				}
		
			# get relevant Tags			
			data_offset = int( dimg.tags['root.ImageList.1.ImageData.Data.Offset'] )
			#data_size = int( dimg.tags['root.ImageList.1.ImageData.Data.Size'] )
			data_type = int( dimg.tags['root.ImageList.1.ImageData.DataType'] )
			im_width = int( dimg.tags['root.ImageList.1.ImageData.Dimensions.0'] )
			im_height = int( dimg.tags['root.ImageList.1.ImageData.Dimensions.1'] )

			self.rawmode = dataTypes[data_type]
			self.mode = 'F'
			self.size = im_width, im_height
		        self.tile = [("raw", (0, 0) + self.size, data_offset, (self.rawmode, 0, 1))]
			# orientation: Whether the first line in the image is the top line on the screen (1), or the bottom line (-1)

	        except:
			raise SyntaxError, "not a valid DM3 file"

Image.register_open("DM3", Dm3ImageFile)
Image.register_extension("DM3", ".dm3")

